/* safeexec.c

   The implementation of an easy to use function for executing an
   external program with the privileges of the real user ID.
   
   Copyright (C) 2008 Tim Stadelmann

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/



/** HEADERS **/

/* autoconf header */
#if HAVE_CONFIG_H
#  include <config.h>
#endif

/* ISO C headers */
#include <stdlib.h>

/* POSIX headers */
#include <unistd.h>
#include <sys/wait.h>

/* program headers */
#include "safeexec.h"


/** FUNCTIONS **/

/* safeexec

   Execute the executable file identified by path with the privileges
   of the real user ID.  Command line arguments are provided in the
   argv array, which is termiated by a NULL entry.  The function waits
   for the executable file to terminate and returns its exit status.
   If dropping the privileges of the effective user ID or executing
   the file failed for any reason, EXIT_FAILURE is returned.  */

int				/* the return status of the executable file */
safeexec (const char *path,	/* the path of the executable file */
	  char *const argv[])	/* the command line arguments */
{
  int result;			/* the return value */
  pid_t pid;			/* the process ID of the child process */

  pid = fork();
  if (pid == 0)
    /* child process */
    {
      if (getuid () != geteuid () )
	{
	  if (setgroups (0, NULL) != 0)
	    exit (EXIT_FAILURE);
	  
	  if (setgid (getgid ()) != 0)
	    exit (EXIT_FAILURE);

	  if (setuid (getuid ()) != 0)
	    exit (EXIT_FAILURE);

	  if (setuid (getuid ()) != 0)
	    exit (EXIT_FAILURE);
	  
	  if (setuid (0) == 0)
	    exit (EXIT_FAILURE);
	}

      execv (path, argv);
      exit (EXIT_FAILURE);	/* only reached if execv fails */
    }
  else
    {
      /* parent process */

      int status;		/* the return status */

      waitpid (pid, &status, 0);
      result = WEXITSTATUS (status);
    }

  return result;
}
